"""Provide micro benchmarking functionality"""

import time
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional


@dataclass
class BenchmarkResults:
    """Benchmark name and associated run times."""

    name: str
    run_times: List[int]


class Benchmark:
    """
    Create and run a microbenchmark.

    Use this class as a context manager. Create the :py:class:`Benchmark` and set the test subject
    with :py:meth:`set_user_code`. Rayne automatically measures and prints the test subject's
    execution time when the context manager exits.

    .. code-block:: python

       with Benchmark() as benchmark:
           benchmark.set_user_code(fibonacci, n=10)

    Attributes:
        name: Rayne assigns a name to each :py:class:`Benchmark`. Rayne uses the name in the output
            to differentiate multiple benchmarks within a single benchmarking script. The default
            name is the string representation of the test subject.

            >>> with Benchmark() as benchmark:
            >>>     benchmark.set_user_code(fibonacci, n=10)
            >>> print(benchmark.name)
            fibonacci

            You can assign a custom name to the :py:class:`Benchmark`.

            .. tab:: Initialization

               >>> with Benchmark(name="Recursive") as benchmark:
               >>>     benchmark.set_user_code(fibonacci, n=10)
               >>> print(benchmark.name)
               Recursive

            .. tab:: Post Initialization

               >>> with Benchmark() as benchmark:
               >>>     benchmark.name = "Recursive"
               >>>     benchmark.set_user_code(fibonacci, n=10)
               >>> print(benchmark.name)
               Recursive

        runs: Rayne runs the test subject several times to build a statistical model of the
            execution time. The :py:attr:`runs` attribute controls how many times Rayne runs the
            test subject. You can modify this value before the :py:class:`Benchmark` context
            manager exits. For example, to run the test subject 2000 times:

            .. tab:: Initialization

               .. code-block:: python

                  with Benchmark(runs=2000) as benchmark:
                      benchmark.subject = fibonacci

            .. tab:: Post Initialization

               .. code-block:: python

                   with Benchmark() as benchmark:
                       benchmark.runs = 2000
                       benchmark.subject = fibonacci
    """

    def __init__(self, name: Optional[str] = None, runs: int = 1000):
        self.name = name
        self.runs = runs
        self.__fut: Optional[Callable] = None  # fut -> function under test
        self.__fut_args: Dict[str, Any]
        self.__run_times: List[int] = []
        self.__clock_latency = 0

    def set_user_code(self, function, **kwargs):
        """
        Set the user code to benchmark.

        Args:
            function (Callable): The benchmark measures the execution time of this function.
            kwargs: The benchmark passes these arguments to ``function``.
        """
        self.__fut = function
        self.__fut_args = kwargs
        if not self.name:
            self.name = function.__name__

    @property
    def run_time(self) -> float:
        """The average execution time, in nanoseconds, of the subject function."""
        if not self.__run_times:
            raise RuntimeError("No user code was measured.")
        return sum(self.__run_times) // self.runs

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self.__fut is None:
            raise RuntimeError("No user code was specified to measure.")
        self.__measure_clock_latency()
        self.__warm_up()
        self.__run_benchmark()
        print(f"{self.name}: {self.run_time} ns")
        return True

    def __measure_clock_latency(self):
        """
        Measure the clock latency

        It takes time to call time.perf_counter_ns(). A benchmark should not include that time
        within; it's not part of executing the function under test. This function measures the
        latency the same number of times as the benchmark runs, and stores the floor of the mean
        latency. Later, the benchmark will subtract this mean latency from each timing measurement.
        """
        total_latency = 0
        for _ in range(0, self.runs):
            start_time = time.perf_counter_ns()
            end_time = time.perf_counter_ns()
            total_latency += end_time - start_time
        self.__clock_latency = total_latency // self.runs

    def __warm_up(self):
        self.__fut(**self.__fut_args)

    def __run_benchmark(self):
        self.__run_times = [0 for _ in range(0, self.runs)]
        for i in range(0, self.runs):
            start_time = time.perf_counter_ns()
            self.__fut(**self.__fut_args)
            end_time = time.perf_counter_ns()
            self.__run_times[i] = max(end_time - start_time - self.__clock_latency, 0)



"""Configuration file for the Sphinx documentation builder."""

import sys
from pathlib import Path

# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# pylint: disable=invalid-name
# cspell: words extlinks furo
# mypy: ignore-errors

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "Rayne"
project_copyright = "2024, brobeson"
author = "brobeson"
version = "0.0.0"
release = "0.0.0"
highlight_language = "python"

sys.path.insert(0, str(Path("..", "rayne").resolve()))

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.extlinks",
    "sphinx.ext.napoleon",
    "sphinx_copybutton",
    "sphinx_inline_tabs",
]

templates_path = ["_templates"]
exclude_patterns = []

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_title = project
html_theme_options = {
    "dark_css_variables": {"color-announcement-background": "#ff5252"},
    # pylint: disable=line-too-long
    "announcement": 'This is alpha software. The interface is subject to breaking changes.<br/>You can see the pre-release road map at the <a href="https://github.com/users/brobeson/projects/6/views/1?filterQuery=milestone%3A%22Next+Release%22">kanban board</a>.',
}

copybutton_exclude = ".linenos, .gp, .go"
extlinks = {"issue": ("https://github.com/brobeson/Rayne/issues/%s", "issue %s")}



"""Unit tests for the Benchmark class."""

# pylint: disable=missing-function-docstring

import pytest
from rayne import Benchmark


def test_benchmark_default_init() -> None:
    b = Benchmark()
    assert b.name is None
    assert b.runs == 1000
    with pytest.raises(RuntimeError):
        assert b.run_time is None


def test_benchmark_custom_init() -> None:
    b = Benchmark(runs=9876, name="Hamlet")
    assert b.name == "Hamlet"
    assert b.runs == 9876
    with pytest.raises(RuntimeError):
        assert b.run_time is None


def test_context() -> None:
    class _RunTest:
        def __init__(self):
            self.__was_run = False

        @property
        def was_run(self) -> bool:
            return self.__was_run

        def __call__(self) -> None:
            self.__was_run = True

    s = _RunTest()
    with Benchmark() as b:
        b.set_user_code(s)
    assert s.was_run
    assert b.run_time > 0


def test_no_subject() -> None:
    with pytest.raises(RuntimeError):
        with Benchmark():
            pass  # Just need to exit the context manager



"""Unit tests for the Fibonacci example functions."""

# pylint: disable=missing-function-docstring

import pytest
from docs.fibonacci import closed_form, recursive

TEST_CASES = [
    (0, 0),
    (1, 1),
    (2, 1),
    (3, 2),
    (4, 3),
    (5, 5),
    (6, 8),
    (7, 13),
    (8, 21),
    (9, 34),
    (10, 55),
    (11, 89),
    (12, 144),
    (13, 233),
    (14, 377),
    (15, 610),
    (16, 987),
    (17, 1597),
    (18, 2584),
    (19, 4181),
]


@pytest.mark.parametrize("n,expected", TEST_CASES)
def test_recursive(n: int, expected: int) -> None:
    assert recursive(n) == expected


@pytest.mark.parametrize("n,expected", TEST_CASES)
def test_closed_form(n: int, expected: int) -> None:
    assert closed_form(n) == expected



"""Example use of Rayne"""

from math import sqrt
from rayne import Benchmark


def closed_form(n: int) -> int:
    """Calculate F(n) using the closed form equation."""
    phi = (1 + sqrt(5)) / 2
    psi = (1 - sqrt(5)) / 2
    return int((phi**n - psi**n) / sqrt(5))


def recursive(n: int) -> int:
    """Calculate F(n) using recursion."""
    if n <= 0:
        return 0
    if n == 1:
        return 1
    return recursive(n - 1) + recursive(n - 2)


if __name__ == "__main__":
    with Benchmark() as benchmark:
        benchmark.set_user_code(closed_form, n=5)

    with Benchmark() as benchmark:
        benchmark.set_user_code(recursive, n=5)



"""Extract milestone details from JSON returned by `gh repo view --json milestones`"""

# pylint: disable=missing-function-docstring

import argparse
import json
import os
import sys

if "CI" not in os.environ:
    sys.exit("Error: You can only run this script within a GitHub workflow.")


def error(title: str, message: str):
    sys.exit(f"::error title={title}::{message}")


def run_exists_cmd():
    milestones = json.loads(arguments.milestones_json)["milestones"]
    version = os.environ["VERSION"]
    for milestone in milestones:
        if milestone["title"] == version:
            with open(os.environ["GITHUB_ENV"], encoding="utf-8", mode="a") as env_file:
                env_file.write(f"MILESTONE={milestone['number']}\n")
            sys.exit(0)
    error(
        "No such milestone",
        f"There is no {version} milestone (https://github.com/brobeson/Rayne/milestones). "
        f"Is it already closed (https://github.com/brobeson/Rayne/milestones?state=closed)?",
    )


def run_issues_cmd():
    issues = json.loads(arguments.issues_json)
    if issues:
        error(
            "Not all issues are done",
            f"Milestone {os.environ['VERSION']} has {len(issues)} open issues "
            f"(https://github.com/brobeson/Rayne/milestone/{os.environ['MILESTONE']}).",
        )
    sys.exit(0)


parser = argparse.ArgumentParser()
subparsers = parser.add_subparsers()
exists_parser = subparsers.add_parser("exists")
exists_parser.add_argument("milestones_json")
exists_parser.set_defaults(func=run_exists_cmd)
issues_parser = subparsers.add_parser("issues")
issues_parser.add_argument("issues_json")
issues_parser.set_defaults(func=run_issues_cmd)
arguments = parser.parse_args()
arguments.func()
